import json
from collections.abc import Iterable

from django.contrib.postgres.utils import prefix_validation_error
from django.contrib.postgres.validators import ArrayMaxLengthValidator
from django.core import checks, exceptions
from django.db.models import Field, Func, Value, lookups
from django.db.models.fields.mixins import CheckFieldDefaultMixin
from django.utils.translation import gettext_lazy as _

from .base import FieldMixin
from .integer import UInt64Field
from .utils import AttributeSetter

__all__ = ["ArrayField"]


class ArrayField(FieldMixin, CheckFieldDefaultMixin, Field):
    nullable_allowed = False
    empty_strings_allowed = False
    default_error_messages = {
        "item_invalid": _("Item %(nth)s in the array did not validate:"),
    }
    _default_hint = ("list", "[]")

    def __init__(self, base_field, size=None, **kwargs):
        self.base_field = base_field
        self.size = size
        if self.size:
            self.default_validators = [
                *self.default_validators,
                ArrayMaxLengthValidator(self.size),
            ]
        # For performance, only add a from_db_value() method if the base field
        # implements it.
        if hasattr(self.base_field, "from_db_value"):
            self.from_db_value = self._from_db_value
        super().__init__(**kwargs)

    def get_internal_type(self):
        return "ArrayField"

    @property
    def model(self):
        try:
            return self.__dict__["model"]
        except KeyError:
            raise AttributeError(
                "'%s' object has no attribute 'model'" % self.__class__.__name__
            )

    @model.setter
    def model(self, model):
        self.__dict__["model"] = model
        self.base_field.model = model

    @classmethod
    def _choices_is_value(cls, value):
        return isinstance(value, (list, tuple)) or super()._choices_is_value(value)

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        if self.base_field.remote_field:
            errors.append(
                checks.Error(
                    "Base field for array cannot be a related field.",
                    obj=self,
                )
            )
        else:
            # Remove the field name checks as they are not needed here.
            base_errors = self.base_field.check()
            if base_errors:
                messages = "\n    ".join(
                    "%s (%s)" % (error.msg, error.id) for error in base_errors
                )
                errors.append(
                    checks.Error(
                        "Base field for array has errors:\n    %s" % messages,
                        obj=self,
                    )
                )
        return errors

    def set_attributes_from_name(self, name):
        super().set_attributes_from_name(name)
        self.base_field.set_attributes_from_name(name)

    @property
    def description(self):
        return "Array of %s" % self.base_field.description

    def db_type(self, connection):
        return "Array(%s)" % self.base_field.db_type(connection)

    def cast_db_type(self, connection):
        return "Array(%s)" % self.base_field.cast_db_type(connection)

    def get_db_prep_value(self, value, connection, prepared=False):
        if isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
            return [
                self.base_field.get_db_prep_value(i, connection, prepared=prepared)
                for i in value
            ]
        return value

    def get_db_prep_save(self, value, connection):
        if isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
            return [self.base_field.get_db_prep_save(i, connection) for i in value]
        return value

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path.startswith("clickhouse_backend.models.array"):
            path = path.replace(
                "clickhouse_backend.models.array", "clickhouse_backend.models"
            )
        kwargs["base_field"] = self.base_field.clone()
        if self.size:
            kwargs["size"] = self.size
        return name, path, args, kwargs

    def to_python(self, value):
        if isinstance(value, str):
            # Assume we're deserializing
            value = json.loads(value)
        if value is None:
            return value

        return [self.base_field.to_python(val) for val in value]

    def _from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return [
            self.base_field.from_db_value(item, expression, connection)
            for item in value
        ]

    def value_to_string(self, obj):
        values = []
        vals = self.value_from_object(obj)
        base_field = self.base_field

        for val in vals:
            if val is None:
                values.append(None)
            else:
                obj = AttributeSetter(base_field.attname, val)
                values.append(base_field.value_to_string(obj))
        return json.dumps(values)

    def get_transform(self, name):
        transform = super().get_transform(name)
        if transform:
            return transform
        # Django always generate "table"."column".size0, which will cause clickhouse exception.
        # DB::Exception: There's no column 'table.column.size0' in table 'table':
        # While processing table.column.size0. (UNKNOWN_IDENTIFIER)
        # if name.startswith("size"):
        #     try:
        #         dimension = int(name[4:])
        #     except ValueError:
        #         pass
        #     else:
        #         return SizeTransformFactory(dimension)

        if "_" not in name:
            try:
                index = int(name)
            except ValueError:
                pass
            else:
                index += 1  # Clickhouse uses 1-indexing
                return IndexTransformFactory(index, self.base_field)
        try:
            start, end = name.split("_")
            start = int(start) + 1
            end = int(end)  # don't add one here because postgres slices are weird
        except ValueError:
            pass
        else:
            return SliceTransformFactory(start, end)

    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        for index, part in enumerate(value):
            try:
                self.base_field.validate(part, model_instance)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["item_invalid"],
                    code="item_invalid",
                    params={"nth": index + 1},
                )

    def run_validators(self, value):
        super().run_validators(value)
        for index, part in enumerate(value):
            try:
                self.base_field.run_validators(part)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["item_invalid"],
                    code="item_invalid",
                    params={"nth": index + 1},
                )


class ArrayRHSMixin:
    def __init__(self, lhs, rhs):
        if isinstance(rhs, (tuple, list)):
            field = lhs.output_field
            expressions = []
            for value in rhs:
                if not hasattr(value, "resolve_expression"):
                    value = Value(field.base_field.get_prep_value(value))
                expressions.append(value)
            rhs = Func(*expressions, function="array")
        super().__init__(lhs, rhs)

    def process_rhs(self, compiler, connection):
        rhs, rhs_params = super().process_rhs(compiler, connection)
        cast_type = self.lhs.output_field.cast_db_type(connection)
        return "%s::%s" % (rhs, cast_type), rhs_params


class ArrayLookup(ArrayRHSMixin, lookups.FieldGetDbPrepValueMixin, lookups.Lookup):
    function = None
    swap_args = False

    def as_clickhouse(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        sql = "%s(%s, %s)" % (
            (self.function, rhs, lhs) if self.swap_args else (self.function, lhs, rhs)
        )
        params = (*lhs_params, *rhs_params)
        return sql, params


@ArrayField.register_lookup
class ArrayContains(ArrayLookup):
    lookup_name = "contains"
    function = "hasAll"


@ArrayField.register_lookup
class ArrayContainedBy(ArrayLookup):
    lookup_name = "contained_by"
    function = "hasAll"
    swap_args = True


@ArrayField.register_lookup
class ArrayExact(ArrayRHSMixin, lookups.Exact):
    pass


@ArrayField.register_lookup
class ArrayOverlap(ArrayLookup):
    lookup_name = "overlap"
    function = "hasAny"


@ArrayField.register_lookup
class ArrayAny(ArrayLookup):
    lookup_name = "any"
    function = "has"

    def process_rhs(self, compiler, connection):
        rhs, rhs_params = super(ArrayRHSMixin, self).process_rhs(compiler, connection)
        cast_type = self.lhs.output_field.base_field.db_type(connection)
        return "%s::%s" % (rhs, cast_type), rhs_params


@ArrayField.register_lookup
class ArrayLenTransform(lookups.Transform):
    lookup_name = "len"
    function = "length"
    output_field = UInt64Field()


class IndexTransform(lookups.Transform):
    def __init__(self, index, base_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = index
        self.base_field = base_field

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        return "%s[%%s]" % lhs, (*params, self.index)

    @property
    def output_field(self):
        return self.base_field


class IndexTransformFactory:
    def __init__(self, index, base_field):
        self.index = index
        self.base_field = base_field

    def __call__(self, *args, **kwargs):
        return IndexTransform(self.index, self.base_field, *args, **kwargs)


class SliceTransform(lookups.Transform):
    def __init__(self, start, end, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if start > end:
            start, end = end, start
        self.offset = start
        self.length = end - start + 1

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        return "arraySlice({}, %s, %s)".format(lhs), (*params, self.offset, self.length)


class SliceTransformFactory:
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __call__(self, *args, **kwargs):
        return SliceTransform(self.start, self.end, *args, **kwargs)


class SizeTransform(lookups.Transform):
    """https://clickhouse.com/docs/en/sql-reference/data-types/array#array-size"""

    def __init__(self, dimension, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dimension = dimension
        self._output_field = UInt64Field()
        while dimension > 0:
            self._output_field = ArrayField(self._output_field)
            dimension -= 1

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        return "%s.size{}".format(self.dimension) % lhs, params

    @property
    def output_field(self):
        return self._output_field


class SizeTransformFactory:
    def __init__(self, dimension):
        self.dimension = dimension

    def __call__(self, *args, **kwargs):
        return SizeTransform(self.dimension, *args, **kwargs)



import collections.abc
import json

from django.contrib.postgres.utils import prefix_validation_error
from django.core import checks, exceptions
from django.db.models import Field, Func, Value, lookups
from django.db.models.fields.mixins import CheckFieldDefaultMixin
from django.utils.translation import gettext_lazy as _

from .array import ArrayField
from .base import FieldMixin
from .integer import UInt64Field
from .utils import AttributeSetter

__all__ = ["MapField"]


class MapField(FieldMixin, CheckFieldDefaultMixin, Field):
    nullable_allowed = False
    empty_strings_allowed = False
    default_error_messages = {
        "key_invalid": _("Key '%(key)s' in the map did not validate:"),
        "value_invalid": _("Value '%(value)s' in the map did not validate:"),
    }
    _default_hint = ("dict", "{}")
    allowed_key_type = {
        "Int8Field",
        "Int16Field",
        "Int32Field",
        "Int64Field",
        "Int128Field",
        "Int256Field",
        "UInt8Field",
        "UInt16Field",
        "UInt32Field",
        "UInt64Field",
        "UInt128Field",
        "UInt256Field",
        "BooleanField",
        "StringField",
        "FixedStringField",
        "UUIDField",
        "ClickhouseDateField",
        "Date32Field",
        "ClickhouseDateTimeField",
        "DateTime64Field",
        "EnumField",
        "Enum8Field",
        "Enum16Field",
        "IPv4Field",
        "IPv6Field",
        "GenericIPAddressField",
    }

    def __init__(self, key_field, value_field, **kwargs):
        self.key_field = key_field
        self.value_field = value_field
        # For performance, only add a from_db_value() method if the base field
        # implements it.
        if hasattr(self.key_field, "from_db_value") or hasattr(
            self.value_field, "from_db_value"
        ):
            self.from_db_value = self._from_db_value
        super().__init__(**kwargs)

    def get_internal_type(self):
        return "MapField"

    @property
    def model(self):
        try:
            return self.__dict__["model"]
        except KeyError:
            raise AttributeError(
                "'%s' object has no attribute 'model'" % self.__class__.__name__
            )

    @model.setter
    def model(self, model):
        self.__dict__["model"] = model
        self.key_field.model = model
        self.value_field.model = model

    @classmethod
    def _choices_is_value(cls, value):
        return isinstance(value, dict) or super()._choices_is_value(value)

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        if self.key_field.remote_field or self.value_field.remote_field:
            errors.append(
                checks.Error(
                    "Key field or value field for map cannot be a related field.",
                    obj=self,
                )
            )
        else:
            key_type = self.key_field.get_internal_type()
            if key_type not in self.allowed_key_type:
                errors.append(
                    checks.Error(
                        "This key field type is invalid.",
                        obj=self,
                    )
                )
            if self.key_field.null:
                errors.append(
                    checks.Error(
                        "Map key must not be null.",
                        obj=self,
                    )
                )
            if getattr(self.key_field, "low_cardinality", False) and key_type not in {
                "StringField",
                "FixedStringField",
            }:
                errors.append(
                    checks.Error(
                        "Only Map key of String and FixedString can be low cardinality.",
                        obj=self,
                    )
                )
            # Remove the field name checks as they are not needed here.
            base_errors = self.key_field.check()
            if base_errors:
                messages = "\n    ".join(
                    "%s (%s)" % (error.msg, error.id) for error in base_errors
                )
                errors.append(
                    checks.Error(
                        "Key field for map has errors:\n    %s" % messages,
                        obj=self,
                    )
                )
            base_errors = self.value_field.check()
            if base_errors:
                messages = "\n    ".join(
                    "%s (%s)" % (error.msg, error.id) for error in base_errors
                )
                errors.append(
                    checks.Error(
                        "Value field for map has errors:\n    %s" % messages,
                        obj=self,
                    )
                )
        return errors

    def set_attributes_from_name(self, name):
        super().set_attributes_from_name(name)
        self.key_field.set_attributes_from_name(name)
        self.value_field.set_attributes_from_name(name)

    @property
    def description(self):
        return "Map of (%s, %s)" % (
            self.key_field.description,
            self.value_field.description,
        )

    def db_type(self, connection):
        return "Map(%s, %s)" % (
            self.key_field.db_type(connection),
            self.value_field.db_type(connection),
        )

    def cast_db_type(self, connection):
        return "Map(%s, %s)" % (
            self.key_field.cast_db_type(connection),
            self.value_field.cast_db_type(connection),
        )

    def get_db_prep_value(self, value, connection, prepared=False):
        if isinstance(value, collections.abc.Mapping):
            return {
                self.key_field.get_db_prep_value(
                    k, connection, prepared=prepared
                ): self.value_field.get_db_prep_value(v, connection, prepared=prepared)
                for k, v in value.items()
            }
        return value

    def get_db_prep_save(self, value, connection):
        if hasattr(value, "as_sql"):
            return value
        if isinstance(value, collections.abc.Mapping):
            return {
                self.key_field.get_db_prep_save(
                    k, connection
                ): self.value_field.get_db_prep_save(v, connection)
                for k, v in value.items()
            }
        return value

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path.startswith("clickhouse_backend.models.map"):
            path = path.replace(
                "clickhouse_backend.models.map", "clickhouse_backend.models"
            )
        kwargs.update(
            {
                "key_field": self.key_field.clone(),
                "value_field": self.value_field.clone(),
            }
        )
        return name, path, args, kwargs

    def to_python(self, value):
        if isinstance(value, str):
            # Assume we're deserializing
            value = json.loads(value)
        if value is None:
            return value
        value = dict(value)
        return {
            self.key_field.to_python(k): self.value_field.to_python(v)
            for k, v in value.items()
        }

    @staticmethod
    def from_db_value_noop(value, expression, connection):
        return value

    def _from_db_value(self, value, expression, connection):
        if value is None:
            return value

        if hasattr(self.key_field, "from_db_value"):
            key_func = self.key_field.from_db_value
        else:
            key_func = self.from_db_value_noop
        if hasattr(self.value_field, "from_db_value"):
            value_func = self.value_field.from_db_value
        else:
            value_func = self.from_db_value_noop
        return {
            key_func(k, expression, connection): value_func(v, expression, connection)
            for k, v in value.items()
        }

    def value_to_string(self, obj):
        values = {}
        vals = self.value_from_object(obj)

        for k, v in vals.items():
            if k is not None:
                obj = AttributeSetter(self.key_field.attname, k)
                k = self.key_field.value_to_string(obj)
            if v is not None:
                obj = AttributeSetter(self.value_field.attname, v)
                v = self.value_field.value_to_string(obj)
            values[k] = v
        return json.dumps(values)

    def get_transform(self, name):
        transform = super().get_transform(name)
        if transform:
            return transform

        return KeyTransformFactory(name, self.key_field, self.value_field)

    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        for key, value in value.items():
            try:
                self.key_field.validate(key, model_instance)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["key_invalid"],
                    code="key_invalid",
                    params={"key": key},
                )
            try:
                self.value_field.validate(value, model_instance)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["value_invalid"],
                    code="value_invalid",
                    params={"value": value},
                )

    def run_validators(self, value):
        super().run_validators(value)
        for key, value in value.items():
            try:
                self.key_field.run_validators(key)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["key_invalid"],
                    code="key_invalid",
                    params={"key": key},
                )
            try:
                self.value_field.run_validators(value)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["value_invalid"],
                    code="value_invalid",
                    params={"value": value},
                )


class MapRHSMixin:
    def __init__(self, lhs, rhs):
        if isinstance(rhs, collections.abc.Mapping):
            expressions = []
            field = lhs.output_field
            for key, value in rhs.items():
                if not hasattr(key, "resolve_expression"):
                    key = Value(field.key_field.get_prep_value(key))
                if not hasattr(value, "resolve_expression"):
                    value = Value(field.value_field.get_prep_value(value))
                expressions.append(key)
                expressions.append(value)
            rhs = Func(*expressions, function="map")
        super().__init__(lhs, rhs)

    def process_rhs(self, compiler, connection):
        rhs, rhs_params = super().process_rhs(compiler, connection)
        cast_type = self.lhs.output_field.cast_db_type(connection)
        return "%s::%s" % (rhs, cast_type), rhs_params


@MapField.register_lookup
class MapHasKey(lookups.FieldGetDbPrepValueMixin, lookups.Lookup):
    lookup_name = "has_key"
    prepare_rhs = False

    def as_clickhouse(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return "mapContains(%s, %s)" % (lhs, rhs), params


@MapField.register_lookup
class MapExact(MapRHSMixin, lookups.Exact):
    pass


@MapField.register_lookup
class MapLenTransform(lookups.Transform):
    lookup_name = "len"
    function = "length"
    output_field = UInt64Field()


@MapField.register_lookup
class KeysTransform(lookups.Transform):
    lookup_name = "keys"
    function = "mapKeys"

    @property
    def output_field(self):
        return ArrayField(self.lhs.output_field.key_field)


@MapField.register_lookup
class ValuesTransform(lookups.Transform):
    lookup_name = "values"
    function = "mapValues"

    @property
    def output_field(self):
        return ArrayField(self.lhs.output_field.value_field)


class KeyTransform(lookups.Transform):
    def __init__(self, key, key_field, value_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key = key
        self.key_field = key_field
        self.value_field = value_field

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        key = self.key_field.get_db_prep_value(self.key, connection)
        return "%s[%%s]" % lhs, (*params, key)

    @property
    def output_field(self):
        return self.value_field


class KeyTransformFactory:
    def __init__(self, key, key_field, value_field):
        self.key = key
        self.key_field = key_field
        self.value_field = value_field

    def __call__(self, *args, **kwargs):
        return KeyTransform(self.key, self.key_field, self.value_field, *args, **kwargs)



from django.apps import apps
from django.db.backends.base.operations import BaseDatabaseOperations
from django.utils.functional import cached_property

from clickhouse_backend import compat
from clickhouse_backend.driver import JSON
from clickhouse_backend.driver.client import insert_pattern
from clickhouse_backend.utils.timezone import get_timezone


class DatabaseOperations(BaseDatabaseOperations):
    compiler_module = "clickhouse_backend.models.sql.compiler"
    cast_char_field_without_max_length = "String"
    integer_field_ranges = {
        # Django fields.
        "SmallIntegerField": (-32768, 32767),
        "IntegerField": (-2147483648, 2147483647),
        "BigIntegerField": (-9223372036854775808, 9223372036854775807),
        "PositiveBigIntegerField": (0, 18446744073709551615),
        "PositiveSmallIntegerField": (0, 65535),
        "PositiveIntegerField": (0, 4294967295),
        "SmallAutoField": (-9223372036854775808, 9223372036854775807),
        "AutoField": (-9223372036854775808, 9223372036854775807),
        "BigAutoField": (-9223372036854775808, 9223372036854775807),
        # Clickhouse fields.
        "Int8Field": (-1 << 7, -1 ^ (-1 << 7)),
        "Int16Field": (-1 << 15, -1 ^ (-1 << 15)),
        "Int32Field": (-1 << 31, -1 ^ (-1 << 31)),
        "Int64Field": (-1 << 63, -1 ^ (-1 << 63)),
        "Int128Field": (-1 << 127, -1 ^ (-1 << 127)),
        "Int256Field": (-1 << 255, -1 ^ (-1 << 255)),
        "UInt8Field": (0, -1 ^ (-1 << 8)),
        "UInt16Field": (0, -1 ^ (-1 << 16)),
        "UInt32Field": (0, -1 ^ (-1 << 32)),
        "UInt64Field": (0, -1 ^ (-1 << 64)),
        "UInt128Field": (0, -1 ^ (-1 << 128)),
        "UInt256Field": (0, -1 ^ (-1 << 256)),
    }
    set_operators = {
        "union": "UNION ALL",
        "intersection": "INTERSECT",
        "difference": "EXCEPT",
    }

    explain_prefix = "EXPLAIN"
    explain_types = {
        "AST",
        "SYNTAX",
        "PLAN",
        "PIPELINE",
        "ESTIMATE",
        "TABLE OVERRIDE",
    }
    explain_settings = {
        "header",
        "description",
        "indexes",
        "actions",
        "json",
    }
    # https://clickhouse.com/docs/en/sql-reference/formats
    supported_output_formats = {
        "TSV",
        "TabSeparated",
        "TabSeparatedRaw",
        "TabSeparatedWithNames",
        "TabSeparatedWithNamesAndTypes",
        "TabSeparatedRawWithNames",
        "TabSeparatedRawWithNamesAndTypes",
        "Template",
        "CSV",
        "CSVWithNames",
        "CSVWithNamesAndTypes",
        "CustomSeparated",
        "CustomSeparatedWithNames",
        "CustomSeparatedWithNamesAndTypes",
        "SQLInsert",
        "Values",
        "Vertical",
        "JSON",
        "JSONStrings",
        "JSONColumns",
        "JSONColumnsWithMetadata",
        "JSONCompact",
        "JSONCompactStrings",
        "JSONCompactColumns",
        "JSONEachRow",
        "JSONEachRowWithProgress",
        "JSONStringsEachRow",
        "JSONStringsEachRowWithProgress",
        "JSONCompactEachRow",
        "JSONCompactEachRowWithNames",
        "JSONCompactEachRowWithNamesAndTypes",
        "JSONCompactStringsEachRow",
        "JSONCompactStringsEachRowWithNames",
        "JSONCompactStringsEachRowWithNamesAndTypes",
        "JSONObjectEachRow",
        "TSKV",
        "Pretty",
        "PrettyNoEscapes",
        "PrettyMonoBlock",
        "PrettyNoEscapesMonoBlock",
        "PrettyCompact",
        "PrettyCompactNoEscapes",
        "PrettyCompactMonoBlock",
        "PrettyCompactNoEscapesMonoBlock",
        "PrettySpace",
        "PrettySpaceNoEscapes",
        "PrettySpaceMonoBlock",
        "PrettySpaceNoEscapesMonoBlock",
        "Prometheus",
        "Protobuf",
        "ProtobufSingle",
        "Avro",
        "Parquet",
        "Arrow",
        "ArrowStream",
        "ORC",
        "RowBinary",
        "RowBinaryWithNames",
        "RowBinaryWithNamesAndTypes",
        "Native",
        "Null",
        "XML",
        "CapnProto",
        "RawBLOB",
        "MsgPack",
    }

    def unification_cast_sql(self, output_field):
        db_type = output_field.db_type(self.connection)
        # normal django fields does not have 'nullable_allowed' attribute
        if (
            not hasattr(output_field, "nullable_allowed")
            or output_field.nullable_allowed
            and not output_field.null
        ):
            db_type = "Nullable(%s)" % db_type
        return "%s::{}".format(db_type)

    def date_extract_sql(self, lookup_type, sql, *args):
        # https://clickhouse.com/docs/en/sql-reference/functions/date-time-functions/
        if lookup_type == "iso_year":
            sql = "toISOYear(%s)" % sql
        elif lookup_type == "day":
            sql = "toDayOfMonth(%s)" % sql
        elif lookup_type == "week":
            sql = "toISOWeek(%s)" % sql
        elif lookup_type == "week_day":
            sql = "modulo(toDayOfWeek(%s), 7) + 1" % sql
        elif lookup_type == "iso_week_day":
            sql = "toDayOfWeek(%s)" % sql
        else:
            sql = "to%s(%s)" % (lookup_type.capitalize(), sql)
        if compat.dj_ge41:
            return sql, args[0]
        else:
            return sql

    def date_trunc_sql(self, lookup_type, sql, *args):
        # https://clickhouse.com/docs/en/sql-reference/functions/date-time-functions#tostartofyear
        *ex, tzname = args
        tzname = tzname or get_timezone()
        sql = "toDate(%s, '%s')" % (sql, tzname)
        if lookup_type != "day":
            sql = f"toStartOf{lookup_type.capitalize()}({sql})"

        if compat.dj_ge41:
            return sql, ex[0]
        else:
            return sql

    def datetime_cast_date_sql(self, sql, *args):
        *ex, tzname = args
        tzname = tzname or get_timezone()
        sql = "toDate(%s, '%s')" % (sql, tzname)
        if compat.dj_ge41:
            return sql, ex[0]
        else:
            return sql

    def datetime_extract_sql(self, lookup_type, sql, *args):
        *ex, tzname = args
        tzname = tzname or get_timezone()
        sql = "toTimeZone(%s, '%s')" % (sql, tzname)
        return self.date_extract_sql(lookup_type, sql, *ex)

    def datetime_trunc_sql(self, lookup_type, sql, *args):
        # https://clickhouse.com/docs/en/sql-reference/functions/date-time-functions/#date_trunc
        *ex, tzname = args
        if tzname:
            sql = "date_trunc('%s', %s, '%s')" % (lookup_type, sql, tzname)
        else:
            sql = "date_trunc('%s', %s)" % (lookup_type, sql)
        if compat.dj_ge41:
            return sql, ex[0]
        else:
            return sql

    def distinct_sql(self, fields, params):
        if fields:
            params = [param for param_list in params for param in param_list]
            return (["DISTINCT ON (%s)" % ", ".join(fields)], params)
        else:
            return ["DISTINCT"], []

    def lookup_cast(self, lookup_type, internal_type=None):
        lookup = "%s"

        # Cast text lookups to text to allow things like filter(x__contains=4)
        if lookup_type in (
            "iexact",
            "contains",
            "icontains",
            "startswith",
            "istartswith",
            "endswith",
            "iendswith",
            "regex",
            "iregex",
        ):
            if internal_type == "IPAddressField" or internal_type == "IPv4Field":
                lookup = "IPv4NumToString(%s)"
            elif internal_type == "IPv6Field":
                lookup = "IPv6NumToString(%s)"
            elif internal_type == "GenericIPAddressField":
                lookup = "replaceRegexpOne(IPv6NumToString(%s), '^::ffff:', '')"
            elif internal_type in ("EnumField", "Enum8Field", "Enum16Field"):
                lookup = "toString(%s)"
            else:
                lookup = "CAST(%s, 'Nullable(String)')"

        if lookup_type == "iexact":
            lookup = "UPPER(%s)" % lookup

        return lookup

    def max_name_length(self):
        """
        Return the maximum length of an identifier.

        https://stackoverflow.com/a/68362429/15096024
        Clickhouse does not have own limits on identifier's length.
        But you're limited by a filesystem's limits, because CH uses filenames as table/column names.
        Ext4 max filename length -- ext4 255 bytes. And a maximum path of 4096 characters.

        An example metadata_path from system.tables:
        /var/lib/clickhouse_backend/store/c13/c13f3d33-7a2e-4e30-813f-3d337a2e7e30/test.sql
        Take off the .sql suffix, actual length limit is 251
        """
        return 251

    def max_in_list_size(self):
        """
        The maximum size of an array is limited to one million elements.
        https://github.com/ClickHouse/ClickHouse/issues/20059
        """
        return 1000000

    def no_limit_value(self):
        return None

    def prepare_sql_script(self, sql):
        return [sql]

    def quote_name(self, name):
        if name.startswith('"') and name.endswith('"'):
            return name  # Quoting once is enough.
        return '"%s"' % name

    def regex_lookup(self, lookup_type):
        if lookup_type == "regex":
            return "match(%s, %s)"
        else:
            return "match(%s, concat('(?i)', %s))"

    @cached_property
    def table_model_dict(self):
        return {model._meta.db_table: model for model in apps.get_models()}

    def sql_flush_table(self, style, table):
        from clickhouse_backend import models

        cluster = None
        if table in self.table_model_dict:
            opts = self.table_model_dict[table]._meta
            engine = getattr(opts, "engine", None)
            if isinstance(engine, models.Distributed):
                table = engine.table
                cluster = engine.cluster

        if cluster:
            return "%s %s ON CLUSTER %s" % (
                style.SQL_KEYWORD("TRUNCATE"),
                style.SQL_FIELD(self.quote_name(table)),
                style.SQL_FIELD(self.quote_name(cluster)),
            )
        return "%s %s" % (
            style.SQL_KEYWORD("TRUNCATE"),
            style.SQL_FIELD(self.quote_name(table)),
        )

    def sql_flush(self, style, tables, *, reset_sequences=False, allow_cascade=False):
        return [self.sql_flush_table(style, table) for table in tables]

    def prep_for_iexact_query(self, x):
        return x

    def bulk_insert_sql(self, fields, placeholder_rows):
        # https://clickhouse-driver.readthedocs.io/en/latest/quickstart.html#inserting-data
        # To insert data efficiently, provide data separately,
        # and end your statement with a VALUES clause:
        return "VALUES"

    def adapt_datefield_value(self, value):
        return value

    def adapt_datetimefield_value(self, value):
        return value

    def adapt_decimalfield_value(self, value, max_digits=None, decimal_places=None):
        return value

    def adapt_json_value(self, value, encoder):
        return JSON(value)

    def explain_query_prefix(self, format=None, **options):
        # bypass normal explain prefix insert in compiler.as_sql
        return ""

    def explain_query(self, format=None, type=None, **settings):
        # https://clickhouse.com/docs/en/sql-reference/statements/explain/
        prefix = self.explain_prefix
        suffix = ""
        options = {}
        unknown_settings = []
        for setting, value in settings.items():
            if setting in self.explain_settings:
                options[setting] = int(bool(value))
            else:
                unknown_settings.append(setting)

        if format:
            supported_formats = self.supported_output_formats
            normalized_format = format.upper()
            if normalized_format not in supported_formats:
                msg = "%s is not a recognized format." % normalized_format
                if supported_formats:
                    msg += " Allowed formats: %s" % ", ".join(sorted(supported_formats))
                else:
                    msg += (
                        f" {self.connection.display_name} does not support any formats."
                    )
                raise ValueError(msg)
            suffix = "FORMAT %s" % normalized_format
        if unknown_settings:
            raise ValueError(
                "Unknown settings: %s" % ", ".join(sorted(unknown_settings))
            )
        if type:
            normalized_type = type.upper()
            if normalized_type not in self.explain_types:
                msg = "%s is not a recognized type. Allowed types: %s." % (
                    normalized_type,
                    ", ".join(sorted(self.explain_types)),
                )
                raise ValueError(msg)
            prefix += " %s" % normalized_type
        if options:
            prefix += " %s" % ", ".join("%s=%s" % i for i in options.items())
        return prefix, suffix

    def last_insert_id(self, cursor, table_name, pk_name):
        query = "SELECT %s FROM %s ORDER BY %s DESC LIMIT 1"
        params = (
            self.quote_name(pk_name),
            self.quote_name(table_name),
            self.quote_name(pk_name),
        )
        cursor.execute(query % params)
        return cursor.fetchone()[0]

    def last_executed_query(self, cursor, sql, params):
        if params:
            if insert_pattern.match(sql):
                return "%s %s" % (sql, ", ".join(map(str, params)))
            else:
                if not isinstance(params, (tuple, dict)):
                    params = tuple(params)
                return sql % params
        return sql

    def settings_sql(self, **kwargs):
        result = []
        params = []
        unknown_settings = []
        for setting, value in kwargs.items():
            if setting in self.connection.introspection.settings:
                result.append("%s=%%s" % setting)
                params.append(value)
            else:
                unknown_settings.append(setting)
        if unknown_settings:
            raise ValueError(
                "Unknown settings: %s" % ", ".join(sorted(unknown_settings))
            )
        sql = "SETTINGS %s" % ", ".join(result)
        return sql, params



import threading

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.utils import DEFAULT_DB_ALIAS
from django.utils.asyncio import async_unsafe
from django.utils.functional import cached_property

from clickhouse_backend import driver as Database  # NOQA

from .client import DatabaseClient
from .creation import DatabaseCreation
from .features import DatabaseFeatures
from .introspection import DatabaseIntrospection
from .operations import DatabaseOperations
from .schema import DatabaseSchemaEditor


class DatabaseWrapper(BaseDatabaseWrapper):
    _connection_sharing_lock = threading.Lock()
    _clickhouse_connections = {}

    vendor = "clickhouse"
    display_name = "ClickHouse"
    # This dictionary maps Field objects to their associated ClickHouse column
    # types, as strings. Column-type strings can contain format strings; they'll
    # be interpolated against the values of Field.__dict__ before being output.
    # If a column type is set to None, it won't be included in the output.
    data_types = {
        # Django fields.
        "SmallAutoField": "Int64",
        "AutoField": "Int64",
        "BigAutoField": "Int64",
        "IPAddressField": "IPv4",
        "GenericIPAddressField": "IPv6",
        "JSONField": "Object('json')",
        "BinaryField": "String",
        "CharField": "FixedString(%(max_length)s)",
        "DateField": "Date32",
        "DateTimeField": "DateTime64(6, 'UTC')" if settings.USE_TZ else "DateTime64(6)",
        "DecimalField": "Decimal(%(max_digits)s, %(decimal_places)s)",
        "FileField": "String",
        "FilePathField": "String",
        "FloatField": "Float64",
        "SmallIntegerField": "Int16",
        "IntegerField": "Int32",
        "BigIntegerField": "Int64",
        "PositiveBigIntegerField": "UInt64",
        "PositiveIntegerField": "UInt32",
        "PositiveSmallIntegerField": "UInt16",
        "SlugField": "String",
        "TextField": "String",
        "UUIDField": "UUID",
        "BoolField": "Bool",
        "BooleanField": "Bool",
        # Clickhouse fields.
        "Int8Field": "Int8",
        "Int16Field": "Int16",
        "Int32Field": "Int32",
        "Int64Field": "Int64",
        "Int128Field": "Int128",
        "Int256Field": "Int256",
        "UInt8Field": "UInt8",
        "UInt16Field": "UInt16",
        "UInt32Field": "UInt32",
        "UInt64Field": "UInt64",
        "UInt128Field": "UInt128",
        "UInt256Field": "UInt256",
        "Float32Field": "Float32",
        "Float64Field": "Float64",
        "StringField": "String",
        "FixedStringField": "FixedString(%(max_bytes)s)",
        "ClickhouseDateField": "Date",
        "Date32Field": "Date32",
        "ClickhouseDateTimeField": "DateTime('UTC')" if settings.USE_TZ else "DateTime",
        "DateTime64Field": "DateTime64(%(precision)s, 'UTC')"
        if settings.USE_TZ
        else "DateTime64(%(precision)s)",
        "EnumField": "Enum",
        "Enum8Field": "Enum8",
        "Enum16Field": "Enum16",
        "ArrayField": "Array",  # Not used by ArrayField.db_type
        "TupleField": "Tuple",  # Not used by TupleField.db_type
        "MapField": "Map",  # Not used by MapField.db_type
        "IPv4Field": "IPv4",
        "IPv6Field": "IPv6",
    }
    low_cardinality_data_types = {
        "Int8Field",
        "Int16Field",
        "Int32Field",
        "Int64Field",
        "Int128Field",
        "Int256Field",
        "UInt8Field",
        "UInt16Field",
        "UInt32Field",
        "UInt64Field",
        "UInt128Field",
        "UInt256Field",
        "Float32Field",
        "Float64Field",
        "BooleanField",
        "StringField",
        "FixedStringField",
        "UUIDField",
        "ClickhouseDateField",
        "Date32Field",
        "ClickhouseDateTimeField",
        "IPv4Field",
        "IPv6Field",
        "GenericIPAddressField",
    }
    operators = {
        "exact": "= %s",
        "iexact": "= UPPER(%s)",
        "contains": "LIKE %s",
        "icontains": "ILIKE %s",
        "gt": "> %s",
        "gte": ">= %s",
        "lt": "< %s",
        "lte": "<= %s",
        "startswith": "LIKE %s",
        "endswith": "LIKE %s",
        "istartswith": "ILIKE %s",
        "iendswith": "ILIKE %s",
    }

    # The patterns below are used to generate SQL pattern lookup clauses when
    # the right-hand side of the lookup isn't a raw string (it might be an expression
    # or the result of a bilateral transformation).
    # In those cases, special characters for LIKE operators (e.g. \, *, _) should be
    # escaped on database side.
    #
    # Note: we use str.format() here for readability as "%" is used as a wildcard for
    # the LIKE operator.
    pattern_esc = r"replaceRegexpAll({}, '\\\\|%%|_', '\\\\\\0')"
    pattern_ops = {
        "contains": "LIKE '%%' || {} || '%%'",
        "icontains": "ILIKE '%%' || {} || '%%'",
        "startswith": "LIKE {} || '%%'",
        "istartswith": "ILIKE {} || '%%'",
        "endswith": "LIKE '%%' || {}",
        "iendswith": "ILIKE '%%' || {}",
    }

    Database = Database
    SchemaEditorClass = DatabaseSchemaEditor
    # Classes instantiated in __init__().
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations

    def __init__(self, settings_dict, alias=DEFAULT_DB_ALIAS):
        super().__init__(settings_dict, alias)
        # Use fake_transaction control whether using fake transaction.
        # Fake transaction is used in test, prevent other database such as postgresql
        # from flush at the end of each testcase. Only use this feature when you are
        # aware of the effect in TransactionTestCase.
        self._fake_transaction = False
        self.migration_cluster = self.settings_dict["OPTIONS"].pop(
            "migration_cluster", None
        )
        self.distributed_migrations = self.settings_dict["OPTIONS"].pop(
            "distributed_migrations", None
        )
        # https://clickhouse-driver.readthedocs.io/en/latest/quickstart.html#streaming-results
        self.max_block_size = self.settings_dict["OPTIONS"].pop("max_block_size", 65409)
        if not self.settings_dict["NAME"]:
            self.settings_dict["NAME"] = "default"

    @property
    def fake_transaction(self):
        return self._fake_transaction

    @fake_transaction.setter
    def fake_transaction(self, value):
        self._fake_transaction = value
        self.features.fake_transaction = self._fake_transaction

    def get_connection_params(self):
        settings_dict = self.settings_dict
        if len(settings_dict["NAME"] or "") > self.ops.max_name_length():
            raise ImproperlyConfigured(
                "The database name '%s' (%d characters) is longer than "
                "Clickhouse's limit of %d characters. Supply a shorter NAME "
                "in settings.DATABASES."
                % (
                    settings_dict["NAME"],
                    len(settings_dict["NAME"]),
                    self.ops.max_name_length(),
                )
            )

        conn_params = {
            "host": settings_dict["HOST"] or "localhost",
            **settings_dict.get("OPTIONS", {}),
        }
        if settings_dict["NAME"]:
            conn_params["database"] = settings_dict["NAME"]
        if settings_dict["USER"]:
            conn_params["user"] = settings_dict["USER"]
        if settings_dict["PASSWORD"]:
            conn_params["password"] = settings_dict["PASSWORD"]
        if settings_dict["PORT"]:
            conn_params["port"] = settings_dict["PORT"]
        return conn_params

    @async_unsafe
    def get_new_connection(self, conn_params):
        # Fix https://github.com/jayvynl/django-clickhouse-backend/issues/53
        with self._connection_sharing_lock:
            if self.alias in self._clickhouse_connections:
                params, conn = self._clickhouse_connections[self.alias]
                if conn_params == params:
                    return conn

            conn = Database.connect(**conn_params)
            self._clickhouse_connections[self.alias] = (conn_params, conn)

        return conn

    def init_connection_state(self):
        pass

    @async_unsafe
    def create_cursor(self, name=None):
        return self.connection.cursor()

    @async_unsafe
    def chunked_cursor(self):
        cursor = self._cursor()
        cursor.cursor.set_stream_results(True, self.max_block_size)
        return cursor

    def set_autocommit(
        self, autocommit, force_begin_transaction_with_broken_autocommit=False
    ):
        self.autocommit = autocommit

    def commit(self):
        pass

    def _savepoint(self, sid):
        pass

    def _savepoint_rollback(self, sid):
        pass

    def _savepoint_commit(self, sid):
        pass

    def _close(self):
        """Close database connection.

        This is a noop, because inner connection is shared between threads.
        """
        pass

    def is_usable(self):
        try:
            # Use a clickhouse_driver cursor directly, bypassing Django's utilities.
            with self.connection.cursor() as cursor:
                cursor.execute("SELECT 1")
        except Database.Error:
            return False
        else:
            return True

    @cached_property
    def ch_version(self):
        with self.temporary_connection() as cursor:
            cursor.execute("SELECT version()")
            row = cursor.fetchone()
        return row[0]

    def get_database_version(self):
        """
        Return a tuple of the database's version.
        E.g. for ch_version "22.9.3.18", return (22, 9, 3, 18).
        """
        return tuple(map(int, self.ch_version.split(".")))



from .fields import *  # noqa: F403
from .fields import __all__ as fields_all
from .functions import *  # noqa: F403
from .functions import __all__ as functions_all
from .migrations import *  # noqa: F403
from .migrations import __all__ as migrations_all

__all__ = [
    "patch_all",
    *fields_all,
    *functions_all,
    *migrations_all,
]


def patch_all():
    patch_functions()  # noqa: F405
    patch_fields()  # noqa: F405
    patch_migrations()  # noqa: F405



from django.db import models

from .json import patch_jsonfield

__all__ = [
    "patch_fields",
    "patch_auto_field",
    "patch_jsonfield",
]


def patch_fields():
    patch_auto_field()
    patch_jsonfield()


def patch_auto_field():
    def rel_db_type_decorator(cls):
        old_func = cls.rel_db_type

        def rel_db_type(self, connection):
            if connection.vendor == "clickhouse":
                return self.db_type(connection)
            return old_func(self, connection)

        cls.rel_db_type = rel_db_type
        return cls

    rel_db_type_decorator(models.AutoField)
    rel_db_type_decorator(models.SmallAutoField)
    rel_db_type_decorator(models.BigAutoField)



from django.db import models


class Func(models.Func):
    @property
    def function(self):
        return self.__class__.__name__

    def deconstruct(self):
        module_name = self.__module__
        name = self.__class__.__name__
        if module_name.startswith("clickhouse_backend.models.functions"):
            module_name = "clickhouse_backend.models"
        return (
            f"{module_name}.{name}",
            self._constructor_args[0],
            self._constructor_args[1],
        )



import re
import typing as T
from contextlib import contextmanager

from clickhouse_driver import connection
from clickhouse_driver.dbapi import connection as dbapi_connection
from clickhouse_driver.dbapi import cursor, errors
from clickhouse_driver.result import IterQueryResult, ProgressQueryResult, QueryResult
from django.conf import settings

from .escape import escape_params
from .pool import ClickhousePool

name_regex = r'"(?:[^"]|\\.)+"'
value_regex = r"(')?(?(1)(?:[^']|\\.)+|\S+)(?(1)'|)"
name_value_regex = f"{name_regex} = {value_regex}"
update_pattern = re.compile(f"^ALTER TABLE ({name_regex}) UPDATE ")


def send_query(self, query, query_id=None, params=None):
    if not self.connected:
        self.connect()

    connection.write_varint(connection.ClientPacketTypes.QUERY, self.fout)

    connection.write_binary_str(query_id or "", self.fout)

    revision = self.server_info.used_revision
    if revision >= connection.defines.DBMS_MIN_REVISION_WITH_CLIENT_INFO:
        client_info = connection.ClientInfo(
            self.client_name, self.context, client_revision=self.client_revision
        )
        client_info.query_kind = connection.ClientInfo.QueryKind.INITIAL_QUERY

        client_info.write(revision, self.fout)

    settings_as_strings = (
        revision
        >= connection.defines.DBMS_MIN_REVISION_WITH_SETTINGS_SERIALIZED_AS_STRINGS
    )
    settings_flags = 0
    if self.settings_is_important:
        settings_flags |= connection.SettingsFlags.IMPORTANT
    connection.write_settings(
        self.context.settings, self.fout, settings_as_strings, settings_flags
    )

    if revision >= connection.defines.DBMS_MIN_REVISION_WITH_INTERSERVER_SECRET:
        connection.write_binary_str("", self.fout)

    connection.write_varint(connection.QueryProcessingStage.COMPLETE, self.fout)
    connection.write_varint(self.compression, self.fout)

    connection.write_binary_str(query, self.fout)

    if revision >= connection.defines.DBMS_MIN_PROTOCOL_VERSION_WITH_PARAMETERS:
        if self.context.client_settings["server_side_params"]:
            # Always settings_as_strings = True
            escaped = escape_params(params or {}, self.context, for_server=True)
        else:
            escaped = {}
        connection.write_settings(
            escaped, self.fout, True, connection.SettingsFlags.CUSTOM
        )

    connection.logger.debug("Query: %s", query)

    self.fout.flush()


# Monkey patch to resolve https://github.com/jayvynl/django-clickhouse-backend/issues/14
connection.Connection.send_query = send_query


class Cursor(cursor.Cursor):
    # Whether to return data in columnar format. For backwards-compatibility,
    # let's default to None.
    columnar = None

    def close(self):
        """Push client back to connection pool"""
        if self.closed:
            return
        self._state = self._states.CURSOR_CLOSED
        self._connection.pool.push(self._client)

    @property
    def closed(self):
        return self._state == self._states.CURSOR_CLOSED

    @property
    def use_numpy(self):
        return self._client.client_settings["use_numpy"]

    @use_numpy.setter
    def use_numpy(self, value):
        self._client.client_settings["use_numpy"] = value
        if value:
            try:
                from clickhouse_driver.numpy.result import (
                    NumpyIterQueryResult,
                    NumpyProgressQueryResult,
                    NumpyQueryResult,
                )

                self._client.query_result_cls = NumpyQueryResult
                self._client.iter_query_result_cls = NumpyIterQueryResult
                self._client.progress_query_result_cls = NumpyProgressQueryResult
            except ImportError as e:
                raise RuntimeError("Extras for NumPy must be installed") from e
        else:
            self._client.query_result_cls = QueryResult
            self._client.iter_query_result_cls = IterQueryResult
            self._client.progress_query_result_cls = ProgressQueryResult

    @contextmanager
    def set_query_execution_args(
        self, columnar: T.Optional[bool] = None, use_numpy: T.Optional[bool] = None
    ):
        original_use_numpy = self.use_numpy
        if use_numpy is not None:
            self.use_numpy = use_numpy

        original_columnar = self.columnar
        if columnar is not None:
            self.columnar = columnar

        yield self

        self.use_numpy = original_use_numpy
        self.columnar = original_columnar

    def __del__(self):
        # If someone forgets calling close method,
        # then release connection when gc happens.
        if not self.closed:
            self.close()

    def _prepare(self):
        """Override clickhouse_driver.Cursor._prepare() to add columnar kwargs.

        See https://github.com/jayvynl/django-clickhouse-backend/issues/119
        """
        execute, execute_kwargs = super()._prepare()
        if self.columnar is not None:
            execute_kwargs["columnar"] = self.columnar
        return execute, execute_kwargs

    def execute(self, operation, parameters=None):
        """fix https://github.com/jayvynl/django-clickhouse-backend/issues/9"""
        if getattr(
            settings, "CLICKHOUSE_ENABLE_UPDATE_ROWCOUNT", True
        ) and update_pattern.match(operation):
            query = self._client.substitute_params(
                operation, parameters, self._client.connection.context
            )
            m = update_pattern.match(query)
            table = m.group(1)
            query_upper = query.upper()
            i = query_upper.rfind(" WHERE ")
            if i > 0:
                j = query_upper.rfind(" SETTINGS ", i + 7)
                if j > 0:
                    where = query[i + 7 : j]
                else:
                    where = query[i + 7 :]
                super().execute(f"select count(*) from {table} where {where}")
                (rowcount,) = self.fetchone()
                self._reset_state()
                self._rowcount = rowcount
        super().execute(operation, parameters)


class Connection(dbapi_connection.Connection):
    """Connection class with support for connection pool."""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("connections_min", 10)
        kwargs.setdefault("connections_max", 100)
        super().__init__(*args, **kwargs)
        self.pool = ClickhousePool(
            dsn=self.dsn,
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
            database=self.database,
            **self.connection_kwargs,
        )

    def _make_client(self):
        """
        :return: a new Client instance.
        """
        return self.pool.pull()

    def close(self):
        self.pool.cleanup()
        self.is_closed = True

    def cursor(self, cursor_factory=Cursor):
        """Use clickhouse_backend.connection.Cursor which support
        connection pool to create cursor."""
        if self.is_closed:
            raise errors.InterfaceError("connection already closed")

        client = self._make_client()
        if self._hosts is None:
            self._hosts = client.connection.hosts
        else:
            client.connection.hosts = self._hosts
        cursor_factory = cursor_factory or Cursor
        return cursor_factory(client, self)



import signal

from django.db.backends.base.client import BaseDatabaseClient


class DatabaseClient(BaseDatabaseClient):
    executable_name = "clickhouse-client"

    @classmethod
    def settings_to_cmd_args_env(cls, settings_dict, parameters):
        args = [cls.executable_name]
        options = settings_dict.get("OPTIONS", {})

        host = settings_dict.get("HOST")
        port = settings_dict.get("PORT")
        dbname = settings_dict.get("NAME")
        user = settings_dict.get("USER")
        passwd = settings_dict.get("PASSWORD")
        secure = options.get("secure")
        dsn = options.get("dsn")

        if dsn:
            args += [dsn]
        else:
            if host:
                args += ["-h", host]
            if port:
                args += ["--port", str(port)]
            if user:
                args += ["-u", user]
            if passwd:
                args += ["--password", passwd]
            if dbname:
                args += ["-d", dbname]
            if secure:
                args += ["--secure"]
        args.extend(parameters)
        return args, None

    def runshell(self, parameters):
        sigint_handler = signal.getsignal(signal.SIGINT)
        try:
            # Allow SIGINT to pass to psql to abort queries.
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            super().runshell(parameters)
        finally:
            # Restore the original SIGINT handler.
            signal.signal(signal.SIGINT, sigint_handler)



from django.db.models import aggregates
from django.db.models.expressions import Star

from clickhouse_backend.models.fields import UInt64Field

__all__ = [
    "uniq",
    "uniqExact",
    "uniqCombined",
    "uniqCombined64",
    "uniqHLL12",
    "uniqTheta",
    "anyLast",
    "argMax",
]


class Aggregate(aggregates.Aggregate):
    @property
    def function(self):
        return self.__class__.__name__

    @property
    def name(self):
        return self.__class__.__name__

    def deconstruct(self):
        module_name = self.__module__
        name = self.__class__.__name__
        if module_name.startswith("clickhouse_backend.models.aggregates"):
            module_name = "clickhouse_backend.models"
        return (
            f"{module_name}.{name}",
            self._constructor_args[0],
            self._constructor_args[1],
        )


class uniq(Aggregate):
    output_field = UInt64Field()
    allow_distinct = True
    empty_result_set_value = 0

    def __init__(self, *expressions, distinct=False, filter=None, **extra):
        expressions = [Star() if exp == "*" else exp for exp in expressions]
        super().__init__(*expressions, distinct=distinct, filter=filter, **extra)


class uniqExact(uniq):
    pass


class uniqCombined(uniq):
    pass


class uniqCombined64(uniq):
    pass


class uniqHLL12(uniq):
    pass


class uniqTheta(uniq):
    pass


class anyLast(Aggregate):
    pass


class argMax(Aggregate):
    arity = 2

    def _resolve_output_field(self):
        return self.get_source_fields()[0]



from .aggregates import *  # noqa: F401,F403
from .aggregates import __all__ as aggregates_all
from .base import ClickhouseModel
from .engines import *  # noqa: F401,F403
from .engines import __all__ as engines_all
from .fields import *  # noqa: F401,F403
from .fields import __all__ as fields_all
from .functions import *  # noqa: F401,F403
from .functions import __all__ as fucntions_all
from .indexes import *  # noqa: F401,F403
from .indexes import __all__ as indexes_all

__all__ = [
    "ClickhouseModel",
    *engines_all,
    *fields_all,
    *fucntions_all,
    *indexes_all,
    *aggregates_all,
]



from .datetime import *  # noqa: F401,F403
from .datetime import __all__ as datetime_all
from .hashes import *  # noqa: F401,F403
from .hashes import __all__ as hashes_all
from .other import *  # noqa: F401,F403
from .other import __all__ as other_all
from .random import *  # noqa: F401,F403
from .random import __all__ as random_all
from .tuples import *  # noqa: F401,F403
from .tuples import __all__ as tuples_all

__all__ = datetime_all + hashes_all + other_all + random_all + tuples_all



import collections
import collections.abc
import copy
import json
from collections.abc import Iterable

from django.contrib.postgres.utils import prefix_validation_error
from django.core import checks, exceptions
from django.db.models import Field, Func, Value, lookups
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from .base import FieldMixin
from .utils import AttributeSetter

__all__ = ["TupleField"]


class TupleField(FieldMixin, Field):
    nullable_allowed = False
    empty_strings_allowed = False
    default_error_messages = {
        "item_invalid": _("Item %(nth)s in the array did not validate:"),
        "value_length_mismatch": _("Value length does not match tuple length."),
    }

    def __init__(self, base_fields, **kwargs):
        self.base_fields = self._check_base_fields(base_fields)
        super().__init__(**kwargs)

    def _check_base_fields(self, base_fields):
        """Check base_fields type when init.

        Because some other init actions depend on correct base_fields type."""

        invalid_error = RuntimeError(
            "'base_fields' must be an iterable containing only(not both) "
            "field instances or (field name, field instance) tuples, "
            "and field name must be valid python identifier."
        )
        if not isinstance(base_fields, Iterable) or isinstance(base_fields, str):
            raise invalid_error

        fields = []
        named_field_flags = []
        for index, field in enumerate(base_fields, start=1):
            if isinstance(field, Field):
                fields.append(field)
                named_field_flags.append(False)
            else:
                try:
                    name, field = field
                except (TypeError, ValueError):
                    raise invalid_error
                if (
                    not isinstance(name, str)
                    or not name.isidentifier()
                    or not isinstance(field, Field)
                ):
                    raise invalid_error
                fields.append((name, field))
                named_field_flags.append(True)

            if field.remote_field:
                raise RuntimeError("Field %ss cannot be a related field." % index)

        if not fields:
            raise RuntimeError("'base_fields' must not be empty.")

        if all(named_field_flags):
            self.is_named_tuple = True
            self._base_fields = tuple(f for _, f in fields)
        elif not any(named_field_flags):
            self.is_named_tuple = False
            self._base_fields = fields
        else:
            raise invalid_error

        # For performance, only add a from_db_value() method if any base field
        # implements it.
        if any(hasattr(field, "from_db_value") for field in self._base_fields):
            self.from_db_value = self._from_db_value
        return fields

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        if errors:
            return errors
        for index, field in enumerate(self._base_fields, 1):
            base_errors = field.check()
            if base_errors:
                messages = "\n    ".join(
                    "%s (%s)" % (error.msg, error.id) for error in base_errors
                )
                return [
                    checks.Error(
                        "Field %ss has errors:\n    %s" % (index, messages), obj=self
                    )
                ]

        return []

    def get_internal_type(self):
        return "TupleField"

    @property
    def model(self):
        try:
            return self.__dict__["model"]
        except KeyError:
            raise AttributeError(
                "'%s' object has no attribute 'model'" % self.__class__.__name__
            )

    @model.setter
    def model(self, model):
        self.__dict__["model"] = model
        for field in self.base_fields:
            if isinstance(field, Field):
                field.model = model
            else:
                field[1].model = model

    @classmethod
    def _choices_is_value(cls, value):
        return isinstance(value, (list, tuple)) or super()._choices_is_value(value)

    def set_attributes_from_name(self, name):
        super().set_attributes_from_name(name)
        for field in self.base_fields:
            if isinstance(field, Field):
                field.set_attributes_from_name(name)
            else:
                field[1].set_attributes_from_name(name)
        if self.is_named_tuple:
            name = self.name
            if name:
                name = name.capitalize()
            else:
                name = "Tuple"
            self.container_class = collections.namedtuple(
                name, (fn for fn, _ in self.base_fields)
            )
        else:
            self.container_class = tuple

    def _convert_type(self, value):
        # When used as Expression output_field
        if not hasattr(self, "container_class"):
            self.set_attributes_from_name("")
        if value is None or isinstance(value, self.container_class):
            return value
        if self.is_named_tuple:
            if isinstance(value, dict):
                return self.container_class(**value)
            return self.container_class(*value)
        # From ClickHouse server 24.8 LTS, tuple("a", "b") returns NamedTuple.
        if isinstance(value, dict):
            return self.container_class(value.values())
        return self.container_class(value)

    @property
    def description(self):
        base_description = ", ".join(field.description for field in self._base_fields)
        return "Tuple of %s" % base_description

    def db_type(self, connection):
        base_type = ", ".join(
            "%s %s" % (field[0], field[1].db_type(connection))
            if self.is_named_tuple
            else field.db_type(connection)
            for field in self.base_fields
        )
        return "Tuple(%s)" % base_type

    def cast_db_type(self, connection):
        base_type = ", ".join(
            "%s %s" % (field[0], field[1].cast_db_type(connection))
            if self.is_named_tuple
            else field.cast_db_type(connection)
            for field in self.base_fields
        )
        return "Tuple(%s)" % base_type

    def call_base_fields(self, func_name, value, *args, **kwargs):
        if value is None:
            return value
        self._validate_length(value)
        values = []
        for i, field in zip(value, self._base_fields):
            values.append(getattr(field, func_name)(i, *args, **kwargs))
        return values

    def get_db_prep_value(self, value, connection, prepared=False):
        if isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
            values = self.call_base_fields(
                "get_db_prep_value", value, connection, prepared=prepared
            )
            return tuple(values)
        return value

    def get_db_prep_save(self, value, connection):
        if isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
            values = self.call_base_fields("get_db_prep_save", value, connection)
            return tuple(values)
        return value

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path.startswith("clickhouse_backend.models.tuple"):
            path = path.replace(
                "clickhouse_backend.models.tuple", "clickhouse_backend.models"
            )
        kwargs.update(
            {
                "base_fields": copy.deepcopy(self.base_fields),
            }
        )
        return name, path, args, kwargs

    def to_python(self, value):
        if isinstance(value, str):
            # Assume we're deserializing
            value = json.loads(value)
        if value is None:
            return value
        value = self.call_base_fields("to_python", value)
        return self._convert_type(value)

    def from_db_value(self, value, expression, connection):
        return self._convert_type(value)

    def _from_db_value(self, value, expression, connection):
        if value is None:
            return value
        self._validate_length(value)
        values = []
        # when set allow_experimental_object_type=1 at session level, value will be a dict.
        if isinstance(value, dict):
            value = value.values()
        for i, field in zip(value, self._base_fields):
            if hasattr(field, "from_db_value"):
                values.append(field.from_db_value(i, expression, connection))
            else:
                values.append(i)
        return self._convert_type(values)

    def value_to_string(self, obj):
        values = []
        vals = self.value_from_object(obj)
        self._validate_length(vals)

        for val, field in zip(vals, self._base_fields):
            if val is None:
                values.append(None)
            else:
                obj = AttributeSetter(field.attname, val)
                values.append(field.value_to_string(obj))
        return json.dumps(values)

    @cached_property
    def base_filed_map(self):
        if self.is_named_tuple:
            result = {}
            for i, (name, field) in enumerate(self.base_fields):
                result[i] = result[name] = field
            return result
        else:
            return dict(enumerate(self.base_fields))

    def get_transform(self, name):
        transform = super().get_transform(name)
        if transform:
            return transform
        try:
            name = int(name)
        except ValueError:
            field = self.base_filed_map.get(name)
        else:
            field = self.base_filed_map.get(name)
            name += 1  # Clickhouse uses 1-indexing
        if not field:
            return None
        return IndexTransformFactory(name, field)

    def _validate_length(self, values):
        if len(values) != len(self.base_fields):
            raise exceptions.ValidationError(
                code="value_length_mismatch",
                message=self.error_messages["value_length_mismatch"],
            )

    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        self._validate_length(value)
        for index, (part, field) in enumerate(zip(value, self._base_fields)):
            try:
                field.validate(part, model_instance)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["item_invalid"],
                    code="item_invalid",
                    params={"nth": index + 1},
                )

    def run_validators(self, value):
        super().run_validators(value)
        self._validate_length(value)
        for index, (part, field) in enumerate(zip(value, self._base_fields)):
            try:
                field.run_validators(part)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["item_invalid"],
                    code="item_invalid",
                    params={"nth": index + 1},
                )


class TupleRHSMixin:
    def __init__(self, lhs, rhs):
        self.empty_qs = False
        if isinstance(rhs, (tuple, list)):
            expressions = []
            # If compare Tuple(Int8, String) to Tuple(Int8), just return empty.
            if len(rhs) != len(lhs.output_field._base_fields):
                self.empty_qs = True
            else:
                try:
                    for value, field in zip(rhs, lhs.output_field._base_fields):
                        if not hasattr(value, "resolve_expression"):
                            value = Value(field.get_prep_value(value))
                        expressions.append(value)
                    rhs = Func(*expressions, function="tuple")
                except exceptions.ValidationError:
                    self.empty_qs = True
        super().__init__(lhs, rhs)

    def process_rhs(self, compiler, connection):
        rhs, rhs_params = super().process_rhs(compiler, connection)
        cast_type = self.lhs.output_field.cast_db_type(connection)
        return "%s::%s" % (rhs, cast_type), rhs_params

    def as_clickhouse(self, compiler, connection):
        if self.empty_qs:
            return "false", []
        return self.as_sql(compiler, connection)


@TupleField.register_lookup
class TupleExact(TupleRHSMixin, lookups.Exact):
    pass


class IndexTransform(lookups.Transform):
    def __init__(self, index, base_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = index
        self.base_field = base_field

    def as_clickhouse(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        params = (*params, self.index)
        return f"tupleElement({lhs}, %s)", params

    @property
    def output_field(self):
        return self.base_field


class IndexTransformFactory:
    def __init__(self, index, base_field):
        self.index = index
        self.base_field = base_field

    def __call__(self, *args, **kwargs):
        return IndexTransform(self.index, self.base_field, *args, **kwargs)



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db.models import fields
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from clickhouse_backend.backend.operations import DatabaseOperations

from .base import FieldMixin

__all__ = [
    "Int8Field",
    "UInt8Field",
    "Int16Field",
    "UInt16Field",
    "Int32Field",
    "UInt32Field",
    "Int64Field",
    "UInt64Field",
    "Int128Field",
    "UInt128Field",
    "Int256Field",
    "UInt256Field",
]


class IntegerFieldMixin(FieldMixin):
    """
    Make positive integer have correct limitation corresponding to clickhouse uint type.
    """

    @cached_property
    def validators(self):
        validators_ = [*self.default_validators, *self._validators]
        internal_type = self.get_internal_type()
        min_value, max_value = DatabaseOperations.integer_field_ranges[internal_type]
        if min_value is not None and not any(
            (
                isinstance(validator, MinValueValidator)
                and (
                    validator.limit_value()
                    if callable(validator.limit_value)
                    else validator.limit_value
                )
                >= min_value
            )
            for validator in validators_
        ):
            validators_.append(MinValueValidator(min_value))
        if max_value is not None and not any(
            (
                isinstance(validator, MaxValueValidator)
                and (
                    validator.limit_value()
                    if callable(validator.limit_value)
                    else validator.limit_value
                )
                <= max_value
            )
            for validator in validators_
        ):
            validators_.append(MaxValueValidator(max_value))
        return validators_

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path.startswith("clickhouse_backend.models.integer"):
            path = path.replace(
                "clickhouse_backend.models.integer", "clickhouse_backend.models"
            )
        return name, path, args, kwargs


class BaseIntegerField(fields.IntegerField):
    def __init__(self, *args, low_cardinality=False, **kwargs):
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)


class Int8Field(IntegerFieldMixin, BaseIntegerField):
    description = _("8bit integer")

    def get_internal_type(self):
        return "Int8Field"


class UInt8Field(IntegerFieldMixin, BaseIntegerField):
    description = _("8bit unsigned integer")

    def get_internal_type(self):
        return "UInt8Field"


class Int16Field(IntegerFieldMixin, BaseIntegerField):
    description = _("16bit integer")

    def get_internal_type(self):
        return "Int16Field"


class UInt16Field(IntegerFieldMixin, BaseIntegerField):
    description = _("16bit unsigned integer")

    def get_internal_type(self):
        return "UInt16Field"


class Int32Field(IntegerFieldMixin, BaseIntegerField):
    description = _("32bit integer")

    def get_internal_type(self):
        return "Int32Field"


class UInt32Field(IntegerFieldMixin, BaseIntegerField):
    description = _("32bit unsigned integer")

    def get_internal_type(self):
        return "UInt32Field"


class Int64Field(IntegerFieldMixin, BaseIntegerField):
    description = _("64bit integer")

    def get_internal_type(self):
        return "Int64Field"


class UInt64Field(IntegerFieldMixin, BaseIntegerField):
    description = _("64bit unsigned integer")

    def get_internal_type(self):
        return "UInt64Field"


class Int128Field(IntegerFieldMixin, BaseIntegerField):
    description = _("128bit integer")

    def get_internal_type(self):
        return "Int128Field"


class UInt128Field(IntegerFieldMixin, BaseIntegerField):
    description = _("128bit unsigned integer")

    def get_internal_type(self):
        return "UInt128Field"


class Int256Field(IntegerFieldMixin, BaseIntegerField):
    description = _("256bit integer")

    def get_internal_type(self):
        return "Int256Field"


class UInt256Field(IntegerFieldMixin, BaseIntegerField):
    description = _("256bit unsigned integer")

    def get_internal_type(self):
        return "UInt256Field"



from django.core.validators import BaseValidator
from django.utils.deconstruct import deconstructible
from django.utils.translation import ngettext_lazy


@deconstructible
class MaxBytesValidator(BaseValidator):
    message = ngettext_lazy(
        "Ensure this value has at most %(limit_value)d byte (it has %(show_value)d).",
        "Ensure this value has at most %(limit_value)d bytes (it has %(show_value)d).",
        "limit_value",
    )
    code = "max_bytes"

    def compare(self, a, b):
        return a > b

    def clean(self, x):
        if isinstance(x, str):
            x = x.encode("utf-8")
        return len(x)
